﻿
// PICマイコン書き込み用の処理
//
// 2011年3月30日作成
// http://kyoutan.jpn.org

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;    // 追加 ファイル操作で使用
using uart; // 追加

namespace WindowsFormsApplication1
{

    public class picctl:uartctl
    {
        public Form1 form1; // PICprogrammerのputmsgが使えるように追加

        // PIC16用のHEXファイルを読んでバッファに格納
        public bool readhexfile16(string filename, uint[] buff)
        {
            string[] str;
            uint buffpos = 0;
            uint offset = 0;

            // バッファを0x3FFFで初期化
            for (int a = 0; a < buff.Length; a++)
            {
                buff[a] = 0x3FFF;
            }

            try
            {
                str = File.ReadAllLines(filename);
            }
            catch
            {
                form1.putmsg("ファイルが開けなかった");
                return false;
            }

            for (int line = 0; line < str.Length; line++)
            {

                //form1.putmsg(str[line]);
                if (str[line][0] == ':')
                {
                    int length, pos = 1, checksum = 0;

                    length = chartobyte(str[line].Substring(pos, 2));

                    // チェックサムのチェック
                    while (true)
                    {
                        int val;
                        if ((length * 2 + 10) < pos) break;
                        val = chartobyte(str[line].Substring(pos, 2));
                        if (val == -1)  // 文字から数への変換でエラー
                        {
                            checksum = 1;
                            break;
                        }
                        checksum += val;
                        pos += 2;
                    }
                    if ((checksum & 0xFF) == 0)
                    {
                        // チェックサムエラーが無いとき
                        offset = (uint)chartobyte(str[line].Substring(3, 2)) << 8;
                        offset += (uint)chartobyte(str[line].Substring(5, 2));

                        //addmsg((address + offset).ToString("X8")+":");

                        buffpos = offset / 2;
                        switch (chartobyte(str[line].Substring(7, 2)))
                        {
                            case 0x00:  // データレコード（2バイトで1ワード）
                                for (pos = 9; pos < (9 + length * 2); )
                                {
                                    uint val;
                                    val = (uint)chartobyte(str[line].Substring(pos, 2)) << 0;
                                    pos += 2;
                                    val += (uint)chartobyte(str[line].Substring(pos, 2)) << 8;
                                    pos += 2;

                                    //addmsg(val.ToString("X8") + " ");

                                    buff[buffpos] = val;
                                    buffpos++;
                                }
                                //putmsg("");
                                break;

                            case 0x01:  // ファイル終了レコード
                                //putmsg("");
                                return true;

                            case 0x02:  // 拡張セグメントアドレスレコード
                            case 0x03:  // スタートセグメントアドレスレコード
                            case 0x04:  // 拡張リニアアドレスレコード（アドレスの上位）
                            case 0x05:  // スタートリニアアドレスレコード
                                form1.putmsg("対応していないHEXファイルでした");
                                return false;
                        }
                    }
                    else
                    {
                        form1.putmsg("チェックサムエラー");
                        return false;
                    }
                }
            }
            form1.putmsg("ファイル終了レコード未確認");
            return true;
        }

        // pic24用のHEXファイルを読んでバッファに格納
        public bool readhexfile32(string filename, uint[] buff)
        {
            string[] str;
            uint buffpos = 0;
            uint address = 0;
            uint offset = 0;

            // バッファを0xFFFFFFで初期化
            for (int a = 0; a < buff.Length; a++)
            {
                buff[a] = 0xFFFFFF;
            }

            try
            {
                str = File.ReadAllLines(filename);
            }
            catch
            {
                form1.putmsg("ファイルが開けなかった");
                return false;
            }

            for (int line = 0; line < str.Length; line++)
            {

                //putmsg(str[line]);
                if (str[line][0] == ':')
                {
                    int length, pos = 1, checksum = 0;

                    length = chartobyte(str[line].Substring(pos, 2));

                    // チェックサムのチェック
                    while (true)
                    {
                        int val;
                        if ((length * 2 + 10) < pos) break;
                        val = chartobyte(str[line].Substring(pos, 2));
                        if (val == -1)  // 文字から数への変換でエラー
                        {
                            checksum = 1;
                            break;
                        }
                        checksum += val;
                        pos += 2;
                    }
                    if ((checksum & 0xFF) == 0)
                    {
                        // チェックサムエラーが無いとき
                        offset = (uint)chartobyte(str[line].Substring(3, 2)) << 8;
                        offset += (uint)chartobyte(str[line].Substring(5, 2));

                        //addmsg((address + offset).ToString("X8")+":");

                        buffpos = (address + offset) / 4;
                        switch (chartobyte(str[line].Substring(7, 2)))
                        {
                            case 0x00:  // データレコード（4バイトで1ワード）
                                for (pos = 9; pos < (length * 2 + 2); )
                                {
                                    uint val;
                                    val = (uint)chartobyte(str[line].Substring(pos, 2)) << 0;
                                    pos += 2;
                                    val += (uint)chartobyte(str[line].Substring(pos, 2)) << 8;
                                    pos += 2;
                                    val += (uint)chartobyte(str[line].Substring(pos, 2)) << 16;
                                    pos += 2;
                                    val += (uint)chartobyte(str[line].Substring(pos, 2)) << 24;
                                    pos += 2;

                                    //addmsg(val.ToString("X8") + " ");

                                    buff[buffpos] = val;
                                    buffpos++;
                                }
                                //putmsg("");
                                break;

                            case 0x01:  // ファイル終了レコード
                                //putmsg("");
                                return true;

                            case 0x04:  // 拡張リニアアドレスレコード（アドレスの上位）
                                address = (uint)chartobyte(str[line].Substring(9, 2)) << 24;
                                address += (uint)chartobyte(str[line].Substring(11, 2)) << 16;
                                //putmsg("");
                                break;

                            case 0x02:  // 拡張セグメントアドレスレコード
                            case 0x03:  // スタートセグメントアドレスレコード
                            case 0x05:  // スタートリニアアドレスレコード
                                form1.putmsg("対応していないHEXファイルでした");
                                return false;
                        }
                    }
                    else
                    {
                        form1.putmsg("チェックサムエラー");
                        return false;
                    }
                }
            }
            form1.putmsg("ファイル終了レコード未確認");
            return true;
        }

        // 1バイト分(2文字)の16進文字列を数に変換
        public int chartobyte(string str)
        {
            int ret;
            if (int.TryParse(str
                , System.Globalization.NumberStyles.HexNumber
                , System.Globalization.CultureInfo.InvariantCulture
                , out ret))
            {
                return ret;
            }
            else
            {
                return -1;
            }
        }


        // メモリー読み出し
        public bool pic24readmemory(uint address, uint size, uint[] buff)
        {
            uint pos = 0;

            //sixserialexe(0x000000);
            sixserialexe(0x040200);
            sixserialexe(0x000000);

            sixserialexe(0x200000 + ((address & 0xFF0000) >> 12));  // MOV #<SourceAddress23:16>, W0
            sixserialexe(0x880190);
            sixserialexe(0x200006 + ((address & 0xFFFF) << 4)); // MOV #<SourceAddress15:0>, W6

            sixserialexe(0x207847);
            sixserialexe(0x000000);

            while (true)
            {
                uint outsize;
                outsize = size - pos;
                if (outsize > 1) outsize = 2;

                pic24regout(pos, outsize, buff);
                pos += outsize;
                if (pos > (size - 1)) break;
                form1.addmsg(".");

                //putmsg(pos.ToString("D"));
            }

            sixserialexe(0x040200);
            sixserialexe(0x000000);

            form1.putmsg("");
            return true;
        }

        // どこからどれだけ書いても大丈夫になった
        public bool pic24writememory(uint address, uint offset, uint size, uint[] buff)
        {
            uint pos;
            pos = offset;
            size += offset;
            while (true)
            {
                uint[] block = new uint[64];
                uint startblock;
                //sixserialexe(0x000000); // NOP このNOPを入れるとなぜか動かない
                sixserialexe(0x040200); // GOTO 0x200
                sixserialexe(0x000000); // NOP

                sixserialexe(0x24001A); // MOV #0x4001, W10
                sixserialexe(0x883B0A); // MOV W10, NVMCON



                // 1ブロック 64ワード初期化
                for (int a = 0; a < 64; a++)
                {
                    block[a] = 0xFFFFFF;
                }

                // 64*2=128=0b1000 0000
                // 0b11111111 11111111 10000000 = 0xFFFF80 でマスクするとブロックの先頭アドレス
                startblock = address & 0xFFFF80;

                uint end = address + ((size - pos) * 2);
                if (end > (startblock + 127)) end = startblock + 127;

                for (; address < end; address += 2)
                {
                    block[(address - startblock) / 2] = buff[pos];
                    pos++;
                }

                form1.addmsg(string.Format("\r\n0x{0:X6} ", startblock));

                sixserialexe(0x200000 + ((startblock & 0xFF0000) >> 12));  // MOV #<SourceAddress23:16>, W0
                sixserialexe(0x880190); // MOV W0, TBLPAG
                sixserialexe(0x200007 + ((startblock & 0xFFFF) << 4)); // MOV #<SourceAddress15:0>, W7

                // 1ブロック 64ワードラッチに書く
                pic24writelatch(block);

                // 書き込み開始
                sixserialexe(0xA8E761);
                sixserialexe(0x000000);
                sixserialexe(0x000000);

                // 書込み終了まで待つ
                while (true)
                {
                    uint ret;

                    sixserialexe(0x040200);
                    sixserialexe(0x000000);
                    sixserialexe(0x803B02);
                    sixserialexe(0x883C22);
                    sixserialexe(0x000000);
                    ret = regoutserialexe();
                    sixserialexe(0x000000);

                    form1.addmsg("+");
                    if ((ret & 8000) == 0) break;
                }
                if (pos > (size - 1)) break;
            }
            sixserialexe(0x040200);
            sixserialexe(0x000000);
            form1.putmsg("");
            return true;
        }

        public bool onebytecmd(byte command)
        {
            byte[] buff = new byte[1];

            DiscardInBuffer();  // 受信バッファクリア
            if (putch(command) == false) return false;
            if (readbyte(buff, 1) == false) return false;
            if (buff[0] == command) return true;
            return false;
        }

        // ライタ接続テスト
        public bool test()
        {
            return onebytecmd(0xAA);
        }

        public bool pic24enticsp()
        {
            return onebytecmd(0x30);
        }

        public bool pic24exiticsp()
        {
            return onebytecmd(0x31);
        }

        public bool sixserialexe(uint code)
        {
            byte command = 0x32;
            bool ret = false;
            byte[] buff = new byte[1];

            DiscardInBuffer();  // 受信バッファクリア
            putch(command);
            putch((byte)((code >> 16) & 0xFF));
            putch((byte)((code >> 8) & 0xFF));
            putch((byte)((code) & 0xFF));
            readbyte(buff, 1);
            if (buff[0] == command) ret = true;
            if (ret == false) form1.putmsg("PIC24送信コマンドエラー");
            return ret;
        }

        public uint regoutserialexe()
        {
            uint ret;
            byte[] buff = new byte[2];

            DiscardInBuffer();  // 受信バッファクリア
            putch(0x33);
            readbyte(buff, 2);
            ret = ((uint)buff[0] << 8) & 0xFF00;
            ret += buff[1];
            return ret;
        }

        // 遅いので専用コマンドで2ワードいっぺんに読むようにしてみたけど、まだ遅い。
        public bool pic24regout(uint offset, uint size, uint[] buff)
        {
            byte[] temp = new byte[6];
            bool ret;
            putch(0x35);
            ret = readbyte(temp, 6);

            if (size > 0)
            {
                buff[offset + 0] = ((uint)temp[3] << 16) + ((uint)temp[0] << 8) + (uint)temp[1];
            }

            if (size > 1)
            {
                buff[offset + 1] = ((uint)temp[2] << 16) + ((uint)temp[4] << 8) + (uint)temp[5];
            }

            return ret;
        }

        // 1ブロック 64ワード ラッチに書く
        // 一部だけ書いただけだと、書いた以外のラッチに以前の値が残るのか、おかしな値が書き込まれる
        // ので1ブロック全部書く
        // 遅いので専用コマンドにして64ワードいっきに転送するようにしてみたけど、まだ遅い。
        public bool pic24writelatch(uint[] buff)
        {
            byte command = 0x34;
            byte[] temp = new byte[1];
            bool ret = false;
            putch(command);
            for (uint a = 0; a < 64; a++)
            {
                putch((byte)((buff[a + 0] & 0x00FF00) >> 8));
                putch((byte)((buff[a + 0] & 0x0000FF) >> 0));
                putch((byte)((buff[a + 1] & 0xFF0000) >> 16));
                putch((byte)((buff[a + 0] & 0xFF0000) >> 16));
                putch((byte)((buff[a + 1] & 0x00FF00) >> 8));
                putch((byte)((buff[a + 1] & 0x0000FF) >> 0));

                a++;

                // ここでライタの応答を待つようにしないと、ライタ側でデータを取りこぼす。
                // 速度低下の原因。
                // ハードウェアフロー制御しているのに取りこぼすってどういうこと？

                // マイコン側でRTSを1にしてもすぐには送信が止まらないのが原因
                // マイコン側の受信バッファを大きくしなきゃ解決できないので
                // このままソフトウェアで同期をとることにする
                readbyte(temp, 1);
                if (temp[0] == command) ret = true;
                else
                {
                    ret = false;
                    break;
                }
                form1.addmsg(".");
            }
            return ret;
        }

        // PIC24接続テスト
        public bool pic24checkconnection()
        {
            uint ret;
            sixserialexe(0x040200);	// GOTO 0x200
            sixserialexe(0x000000);	// NOP
            sixserialexe(0x255AA2);	// MOV #0x55AA,W2
            sixserialexe(0x883C22); // MOV W2,VISI
            sixserialexe(0x000000);	// NOP
            ret = regoutserialexe();   // read VISI reg
            sixserialexe(0x000000);	// NOP

            if (ret == 0x55AA) return true;
            else return false;
        }

        // PIC24 チップイレース(ユーザーメモリ)
        public bool pic24chiperase()
        {
            //sixserialexe(0x000000);
            sixserialexe(0x040200);
            sixserialexe(0x000000);

            sixserialexe(0x2404FA);
            sixserialexe(0x883B0A);

            sixserialexe(0x200000); // MOV #<PAGEVAL>, W0 (W0に0x80以上を入れると、ユーザーメモリとコンフィギュレーションメモリの両方を消去)
            sixserialexe(0x880190);
            sixserialexe(0x200000);
            sixserialexe(0xBB0800);
            sixserialexe(0x000000);
            sixserialexe(0x000000);

            sixserialexe(0xA8E761);
            sixserialexe(0x000000);
            sixserialexe(0x000000);

            // NVMCON の bit15(WR)が0になるまで待つ
            while (true)
            {
                uint ret;
                sixserialexe(0x040200);
                sixserialexe(0x000000);
                sixserialexe(0x803B02);
                sixserialexe(0x883C22);
                sixserialexe(0x000000);
                ret = regoutserialexe();
                sixserialexe(0x000000);
                if ((ret & 0x8000) == 0) break;
                form1.addmsg(".");
            }

            form1.putmsg("");
            return true;
        }

        public bool pic16enticsp()
        {
            return onebytecmd(0x01);
        }

        public bool pic16exticsp()
        {
            return onebytecmd(0x31);
        }

        public bool pic16loadconfig()
        {
            return onebytecmd(0x02);
        }

        public bool pic16loadpgm(uint code)
        {
            byte command = 0x03;
            bool ret = false;
            byte[] buff = new byte[1];

            DiscardInBuffer();  // 受信バッファクリア
            putch(command);
            putch((byte)((code >> 8) & 0xFF));
            putch((byte)((code) & 0xFF));
            readbyte(buff, 1);
            if (buff[0] == command) ret = true;
            return ret;
        }

        public uint pic16readpgm()
        {
            uint ret;
            byte[] buff = new byte[2];

            DiscardInBuffer();  // 受信バッファクリア
            putch(0x04);
            readbyte(buff, 2);
            ret = ((uint)buff[0] << 8) & 0xFF00;
            ret += buff[1];
            return ret;
        }

        public bool pic16incaddress()
        {
            return onebytecmd(0x05);
        }

        public bool pic16beginerase()
        {
            return onebytecmd(0x06);
        }

        public bool pic16beginprogrammingonly()
        {
            return onebytecmd(0x07);
        }

        public bool pic16bulkerasepgm()
        {
            return onebytecmd(0x08);
        }

        public bool pic16bulkerasedat()
        {
            return onebytecmd(0x09);
        }

        public bool pic16chiperase()
        {
            return onebytecmd(0x0A);
        }

        public bool pic16loaddat()
        {
            return onebytecmd(0x0B);
        }

        public uint pic16readdat()
        {
            uint ret;
            byte[] buff = new byte[1];

            DiscardInBuffer();  // 受信バッファクリア
            putch(0x0C);
            readbyte(buff, 1);
            ret = buff[0];
            return ret;
        }

    }
}
